import numpy as np
from copy import deepcopy
from qepy.driver import Driver
from qepy.io import QEInput
from qepy import constants
from ase.calculators.calculator import Calculator
from ase.units import create_units
from ase.geometry import wrap_positions
import ase.io

units = create_units('2006').copy()
units['Bohr'] = constants.BOHR_RADIUS_SI * 1E10
units['Ry'] = constants.RYTOEV

class QEpyCalculator(Calculator):
    """QEpy Calculator for ASE.

    QEpy initialization depends on the input file. Here are some ways to generate input file :

         - Give a file `inputfile` or/and a dict `qe_options` will generate a temporary input file with new atomic information.
         - Give input file name `inputfile` and `from_file = True` will read the structures to atoms (ase.Atoms) from `inputfile` and initialize the QEpy with `inputfile`.
         - Give dict `ase_espresso` will generate a temporary input file with `ase.io.espresso.write_espresso_in <https://wiki.fysik.dtu.dk/ase/ase/io/formatoptions.html#ase.io.espresso.write_espresso_in>`__ (Only works for program PW).


    Parameters
    ----------
    atoms : ase.Atoms
        ASE atoms object
    inputfile : str
        Name of QE input file
    from_file : str
        Read the structure from inputfile, and use the inputfile for the QE Initialization.
    wrap : bool
        If True, wrap the atoms back to cell before update the ions.
    extrapolation : str or bool
        If False, every step will re-run the QE from file.
    ase_espresso : dict
        A dictionary with some parameters to generate PW input.
    qe_options: dict
        A dictionary with input parameters for QE to generate QE input file.
    comm : object
        Parallel communicator
    ldescf : bool
        If True, also print the scf correction term in the iteratively mode
    iterative : bool
        Iteratively run the scf or tddft
    task : str
        Task of the driver :

          - 'scf' : scf task
          - 'optical' : Optical absorption spectrum (TDDFT)
          - 'nscf' : read file from scf

    embed : object (Fortran)
        embed object for QE
    prefix : str
        prefix of QE input/output
    outdir : str
        The output directory of QE
    logfile : str, file or bool
        The screen output of QE. If it is a file descriptor, it should open with 'w+'.

          - None : Show on the screen.
          - str  : Save to the given file.
          - True : Save to the temporary file, see also :func:`qepy.driver.Driver.get_output`.


    .. note::
         - if `from_file` is True, the qe_options will not use.

         - The bool `gather` parameter in functions means the data is gathered in root or distributed in all cpus.
         - if `ase_espresso` is set, will use ase module to write QE input file. `ase_espresso` contains :

             + input_data: dict
             + pseudopotentials: dict
             + kpts: (int, int, int) or dict
             + koffset: (int, int, int)
             + crystal_coordinates: bool

            More details, see `ase.io.espresso.write_espresso_in <https://wiki.fysik.dtu.dk/ase/ase/io/formatoptions.html#ase.io.espresso.write_espresso_in>`__.

    """

    implemented_properties=['energy', 'forces', 'stress']

    def __init__(self, atoms = None, inputfile = None, from_file = False, wrap = False, extrapolation = True,
            ase_espresso = None, qe_options = None, comm = None, ldescf = False, iterative = False,
            task = 'scf', embed = None, prefix = None, outdir = None, logfile = None, prog = 'pw', **kwargs):
        Calculator.__init__(self, atoms = atoms, **kwargs)
        #
        self.lstart = False
        self.optimizer = None
        #
        self.atoms = atoms
        self.inputfile = inputfile
        self.wrap = wrap
        self.extrapolation = extrapolation
        self.from_file = from_file
        self.ase_espresso = ase_espresso
        self.prog = prog
        self.qe_options = qe_options
        #
        if self.inputfile is not None and self.atoms is None :
            self.atoms = ase.io.read(self.inputfile, format='espresso-in')
        if self.atoms is not None :
            self.atoms.calc = self
        #
        if self.from_file :
            self.basefile = None
        else :
            self.basefile = self.inputfile
            self.inputfile = 'qepy_input_tmp.in'
        #
        self.atoms_save = None
        self.driver = None
        self.iter = 0
        self.qeinput = QEInput()
        #
        self.qepy_options = {
                'comm' : comm,
                'ldescf' : ldescf,
                'iterative' : iterative,
                'task' : task,
                'embed' : embed,
                'logfile' : logfile,
                }
        self.qepy_options.update(kwargs)
        #
        self.parameters['qe_options'] = qe_options
        self.parameters['prog'] = prog
        self.parameters['kpts'] = None
        #
        self.restart()

    def set_atoms(self, atoms):
        atoms.pbc = [True, True, True]
        self.atoms = atoms

    @property
    def qe_options(self):
        return self._qe_options

    @qe_options.setter
    def qe_options(self, value):
        if value is not None :
            # avoid changing the input value
            value = deepcopy(value)
        self._qe_options = value

    def restart_driver(self, qe_options=None, inputfile=None, atoms=None):
        if qe_options is not None :
            self.from_file = False
            self.qe_options = qe_options
        if inputfile is not None :
            self.from_file = True
            self.inputfile = inputfile
        if atoms is not None :
            self.atoms_save = atoms.copy()
            self.atoms = atoms
        if self.lstart :
            self.stop()
        self.driver_initialise()

    def restart(self):
        self.results = {}

    def driver_initialise(self, atoms = None, **kwargs):
        atoms = atoms or self.atoms
        if self.wrap : atoms.wrap()
        if not self.from_file :
            if self.ase_espresso :
                ase.io.write(self.inputfile, atoms, format = 'espresso-in', **self.ase_espresso)
            else :
                self.qeinput.write_qe_input(self.inputfile, atoms=atoms, basefile=self.basefile,
                        qe_options=self.qe_options, prog=self.prog)
        self.driver = Driver(self.inputfile, **self.qepy_options)
        self.lstart = True

    def update_atoms(self, atoms = None, **kwargs):
        atoms = atoms or self.atoms
        #
        if not self.extrapolation :
            self.stop()
            self.driver_initialise(atoms = atoms)
            return
        #
        if self.wrap :
            positions = wrap_positions(atoms.positions, atoms.cell)
        else :
            positions = atoms.positions

        positions = positions /units['Bohr']
        #
        lattice = atoms.cell /units['Bohr']
        lattice_old =self.driver.get_ions_lattice()
        if np.allclose(lattice, lattice_old): lattice = None
        #
        self.driver.update_ions(positions=positions, lattice=lattice, **kwargs)

    def check_restart(self, atoms = None):
        restart = True

        if atoms is None or (self.atoms_save and atoms == self.atoms_save):
            restart = False

        if self.iter == 0 :
            restart = True

        if restart :
            if self.atoms is None :
                self.atoms = atoms
                self.atoms.calc = self
            if atoms is not None :
                self.atoms_save = atoms.copy()
            else :
                atoms = self.atoms
            self.restart()
            if self.iter > 0 :
                self.update_atoms(atoms)
            else :
                self.driver_initialise(atoms = atoms)

        return restart

    def update_optimizer(self, atoms = None, **kwargs):
        if self.check_restart(atoms):
            restart = True
            self.iter += 1
            self.driver.scf(**kwargs)
        else :
            restart = False
        return restart

    def stop(self, what = 'all', **kwargs):
        """See :func:`qepy.driver.Driver.stop`"""
        if not self.lstart :
            pass
        else :
            self.lstart = False
            return self.driver.stop(what = 'all', **kwargs)

    @property
    def is_root(self):
        """See :func:`qepy.driver.Driver.is_root`"""
        return self.driver.is_root

    def get_density(self, gather = True):
        """See :func:`qepy.driver.Driver.get_density`"""
        return self.driver.get_density(gather=gather) / (units['Bohr'] ** 3)

    def get_wave_function(self, band=None, kpt=0):
        """See :func:`qepy.driver.Driver.get_wave_function`"""
        return self.driver.get_wave_function(band=band, kpt=kpt)

    def get_number_of_k_points(self):
        """See :func:`qepy.driver.Driver.get_number_of_k_points`"""
        return self.driver.get_number_of_k_points()

    def get_potential_energy(self, atoms = None, **kwargs):
        """See :func:`qepy.driver.Driver.get_potential_energy`"""
        self.update_optimizer(atoms)
        return self.driver.get_energy(**kwargs) * units['Ry']

    def get_forces(self, atoms = None, icalc = 0):
        """See :func:`qepy.driver.Driver.get_forces`"""
        if self.update_optimizer(atoms) or self._forces is None:
            self._forces = self.driver.get_forces(icalc=icalc)* units['Ry'] / units['Bohr']
        return self._forces

    def get_dos(self, qe_options = None, spin = None, inputfile = None, **kwargs):
        """ calculate density of states (DOS) and return the tuple (energies, dos)

        Parameters
        ----------
        qe_options : dict
            A dictionary with input parameters for QE to generate QE input file.
        spin : int

              - None : total DOS
              - 0 : up
              - 1 : down

        inputfile : str
            Directly use the inputfile to run the QE calculation.
        """
        if qe_options or inputfile :
            if qe_options is not None :
                qe_options = deepcopy(qe_options)
                qe_options_dos = {
                        '&control' : {'calculation': "'nscf'"},
                        '&system' : {'occupations': "'tetrahedra'"},
                        }
                qe_options = QEInput.update_options(options=qe_options_dos, qe_options=qe_options)
            self.restart_driver(qe_options=qe_options, inputfile=inputfile)
            self.driver.non_scf()

        from ase.dft.dos import DOS
        dos = DOS(self, **kwargs)
        energies = dos.get_energies()
        dos = dos.get_dos(spin=spin)
        return energies, dos

    def get_band_structure(self, qe_options = None, kpts = None, reference = None, inputfile = None, **kwargs):
        """ calculate band structure and return ase BandStructure.

        Parameters
        ----------
        qe_options : dict
            A dictionary with input parameters for QE to generate QE input file.
        kpts : ase.dft.kpoints.bandpath
            A band path
        reference : float
            Fermi level which should come from the SCF or DOS calculation.
            If not set, will try to use previous calculation instead of band structure calculation.
        inputfile : str
            Directly use the inputfile to run the QE calculation.
        """
        from ase.calculators.calculator import kpts2ndarray
        from ase.spectrum.band_structure import get_band_structure

        if reference is None and self.lstart :
            reference = self.get_fermi_level()

        if qe_options is not None and kpts is not None :
            qe_options = deepcopy(qe_options)
            kpoints = []
            kgrid = kpts2ndarray(kpts, atoms=self.atoms)
            kpoints.append(len(kgrid))
            w = 1.0/len(kgrid)
            for k in kgrid:
                kpoints.append(f'{k[0]:.10f} {k[1]:.10f} {k[2]:.10f} {w}')
            qe_options_bands = {
                    '&control' : {'calculation': "'bands'"},
                    'k_points crystal_b' : kpoints
                    }
            qe_options = QEInput.update_options(options=qe_options_bands, qe_options=qe_options)

        self.restart_driver(qe_options=qe_options, inputfile=inputfile)
        self.driver.non_scf()

        band = get_band_structure(calc = self, path = kpts, reference=reference)
        return band

    def get_stress(self, atoms = None):
        """Return the stress tensor in the Voigt order (xx, yy, zz, yz, xz, xy)"""
        if self.update_optimizer(atoms) or self._stress is None:
            stress = self.driver.get_stress()
            self._stress = np.array(
                [stress[0, 0], stress[1, 1], stress[2, 2],
                 stress[1, 2], stress[0, 2], stress[0, 1]]) * -1 * units['Ry'] / (units['Bohr'] ** 3)
        return self._stress

    def get_number_of_bands(self):
        """See :func:`qepy.driver.Driver.get_number_of_bands`"""
        return self.driver.get_number_of_bands()

    def get_xc_functional(self):
        """See :func:`qepy.driver.Driver.get_xc_functional`"""
        return self.driver.get_xc_functional()

    def get_bz_k_points(self):
        """See :func:`qepy.driver.Driver.get_bz_k_points`"""
        return self.driver.get_bz_k_points()

    def get_bz_to_ibz_map(self):
        return None

    def get_number_of_spins(self):
        """See :func:`qepy.driver.Driver.get_number_of_spins`"""
        return self.driver.get_number_of_spins()

    def get_spin_polarized(self):
        """See :func:`qepy.driver.Driver.get_spin_polarized`"""
        return self.driver.get_spin_polarized()

    def get_ibz_k_points(self):
        """See :func:`qepy.driver.Driver.get_ibz_k_points`"""
        return self.driver.get_ibz_k_points()

    def get_k_point_weights(self):
        """See :func:`qepy.driver.Driver.get_k_point_weights`"""
        return self.driver.get_k_point_weights()

    def get_pseudo_density(self, spin=None, pad=True, gather = False):
        """See :func:`qepy.driver.Driver.get_pseudo_density`"""
        return self.driver.get_pseudo_density(spin=spin, pad=pad, gather=gather) / (units['Bohr'] ** 3)

    def get_effective_potential(self, spin=0, pad=True):
        """See :func:`qepy.driver.Driver.get_effective_potential`"""
        return self.driver.get_effective_potential(spin=spin, pad=pad)

    def get_pseudo_wave_function(self, band=None, kpt=0, spin=0, broadcast=True,
                                 pad=True):
        """See :func:`qepy.driver.Driver.get_pseudo_wave_function`"""
        return self.driver.get_pseudo_wave_function(band=band, kpt=kpt, spin=spin, broadcast=broadcast, pad=pad)

    def get_eigenvalues(self, kpt=0, spin=0):
        """See :func:`qepy.driver.Driver.get_eigenvalues`"""
        return self.driver.get_eigenvalues(kpt=kpt, spin=spin) * units['Ry']

    def get_occupation_numbers(self, kpt=0, spin=0):
        """See :func:`qepy.driver.Driver.get_occupation_numbers`"""
        return self.driver.get_occupation_numbers(kpt=kpt, spin=spin)

    def get_fermi_level(self):
        """See :func:`qepy.driver.Driver.get_fermi_level`"""
        return self.driver.get_fermi_level() * units['Ry']

    # def initial_wannier(self, initialwannier, kpointgrid, fixedstates,
                        # edf, spin, nbands):
        # """See :func:`qepy.driver.Driver.initial_wannier`"""
        # raise NotImplementedError

    # def get_wannier_localization_matrix(self, nbands, dirG, kpoint,
                                        # nextkpoint, G_I, spin):
        # """See :func:`qepy.driver.Driver.get_wannier_localization_matrix`"""
        # raise NotImplementedError

    def get_magnetic_moment(self, atoms=None):
        """See :func:`qepy.driver.Driver.get_magnetic_moment`"""
        return self.driver.get_magnetic_moment(atoms=atoms)

    def get_number_of_grid_points(self, gather = True):
        """See :func:`qepy.driver.Driver.get_number_of_grid_points`"""
        return self.driver.get_number_of_grid_points(gather=gather)

    @staticmethod
    def get_atoms_from_qepy():
        """Return the atom.Atoms from QE."""
        atoms = Driver.get_ase_atoms()
        return atoms



import os
import sys
import types
from functools import wraps


class Logger(type):
    def __new__(cls, name, bases, attrs):

        for attr_name, attr_value in attrs.items():
            if isinstance(attr_value, types.FunctionType):
                attrs[attr_name] = cls.stdout2file(attr_value)

        return super().__new__(cls, name, bases, attrs)

    @staticmethod
    def stdout2file(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if hasattr(args[0], 'fileobj'):
                fobj = args[0].fileobj
            else :
                fobj = None
            stdout = None
            if fobj is not None :
                if os.fstat(1).st_ino != os.fstat(fobj.fileno()).st_ino :
                    stdout = os.dup(1)
                    os.dup2(fobj.fileno(), 1)
            results = function(*args, **kwargs)
            if stdout is not None :
                os.dup2(stdout, 1)
                os.close(stdout)
            return results
        return wrapper

def stdout2file(fileobj = None):
    def decorator(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            if hasattr(args[0], 'fileobj'):
                fobj = args[0].fileobj
            else :
                fobj = None
            stdout = None
            if fobj is not None :
                if os.fstat(1).st_ino != os.fstat(fobj.fileno()).st_ino :
                    stdout = os.dup(1)
                    os.dup2(fobj.fileno(), 1)
            results = function(*args, **kwargs)
            if stdout is not None :
                os.dup2(stdout, 1)
                os.close(stdout)
            return results
        return wrapper
    return decorator


env = {
    'STDOUT' : sys.stdout,  # file descriptor of sprint
    'DRIVER' : None, # save the instance of driver class
}



from qepy.driver import Driver
import numpy as np
import pathlib

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'

def test_0_scf():
    # Run scf
    driver = Driver(inputfile, comm)
    driver.scf()
    converged = driver.check_convergence()
    energy = driver.get_energy()
    driver.save()
    #
    energy = driver.get_energy()
    if driver.is_root :
        print('converged :\n', converged)
        print('energy :\n', energy)
    assert converged
    assert np.isclose(energy, -552.93477389, atol = 1E-6)

def test_1_tddft_continue():
    # Run TDDFT after scf, without stop
    driver = Driver(inputfile, comm, task = 'optical', progress = True)
    driver.scf()
    dipole = driver.get_dipole_tddft()
    if driver.is_root :
        print('dipople:\n', dipole)
    assert np.isclose(dipole[0, 0], 0.56199, atol = 1E-3)
    driver.stop()

def test_2_tddft_iterative():
    # Run TDDFT
    driver = Driver(inputfile, comm, task = 'optical', iterative = True)
    for i in range(5):
        driver.diagonalize()
    dipole = driver.get_dipole_tddft()
    if driver.is_root :
        print('dipople:\n', dipole)
    assert np.isclose(dipole[0, 0], 0.54355, atol = 1E-3)
    driver.stop()

def test_3_tddft_restart():
    driver = Driver(inputfile, comm, task = 'optical', iterative = True)
    # restart from 5
    driver.tddft_restart(istep=5)
    for i in range(5):
        driver.diagonalize()
    dipole = driver.get_dipole_tddft()
    if driver.is_root :
        print('dipople:\n', dipole)
    assert np.isclose(dipole[0, 0], 0.56199, atol = 1E-3)
    driver.stop()



import qepy
import unittest
import pathlib
import shutil
try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_fake.in'


class Test(unittest.TestCase):
    def test_pw_comm(self):
        # return # Take some time to finish
        for i in range(1000): # Error happened ~450
            qepy.qepy_pwscf(inputfile, commf)
            # qepy.qepy_mod.qepy_set_stdout('/dev/null')
            qepy.qepy_mod.qepy_set_stdout('al.out')
            qepy.qepy_stop_run(0, what = 'no')
            qepy.qepy_mod.qepy_write_stdout(f'Test step = {i}')
            print('istep = ', i, flush = True)

    def test_initial_comm(self):
        return # read_file take some time
        inputobj = qepy.qepy_common.input_base()
        inputobj.prefix = 'tmp'
        if commf : inputobj.my_world_comm = commf
        for i in range(50000): # Error happened ~16000
            # qepy.qepy_mod.qepy_set_stdout('/dev/null')
            qepy.qepy_mod.qepy_set_stdout('al.out')
            qepy.qepy_initial(inputobj)
            # qepy.read_file()
            qepy.qepy_stop_run(0, what = 'no')
            qepy.qepy_mod.qepy_write_stdout(f'Test step = {i}')

    def tearDown(self):
        if comm and comm.rank == 0 :
            path = pathlib.Path('.')
            for f in path.glob('al.*'):
                if f.is_file():
                    f.unlink()
                else :
                    shutil.rmtree(f)


if __name__ == "__main__":
    unittest.main()



import qepy
try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    comm = comm.py2f()
except Exception:
    comm = None

oldxml = False # oldxml version QE

fname = 'qe_in.in'
qepy.qepy_pwscf(fname, comm, oldxml)

embed = qepy.qepy_common.embed_base()

if oldxml :
    qepy.oldxml_pw_restart.pw_readfile('header')
    qepy.oldxml_pw_restart.pw_readfile('reset')
    qepy.oldxml_pw_restart.pw_readfile('dim')
    qepy.oldxml_pw_restart.pw_readfile('bs')
    if qepy.basis.get_starting_pot().strip() != 'file' :
        qepy.oldxml_potinit(starting = 'file')
    if qepy.basis.get_starting_wfc().strip() != 'file' :
        qepy.oldxml_wfcinit(starting = 'file')
else :
    qepy.qepy_pw_restart_new.qepy_read_xml_file(alloc=False)
    if qepy.basis.get_starting_pot().strip() != 'file' :
        qepy.qepy_potinit(starting = 'file')
    if qepy.basis.get_starting_wfc().strip() != 'file' :
        qepy.qepy_wfcinit(starting = 'file')

energy = qepy.qepy_calc_energies(embed)
qepy.qepy_stop_run(0, what = 'no')



from collections import OrderedDict

class QEInput(object):
    """Input for QE

    Read/write the QE input file. The input `qe_options` or the return of reading is a dictionary.
    There are two main types of QE input parameters : namelist and card.
    In the `qe_options`, for namelist, the key is always start with '&',  and the value is a dictionary.
    But for card, the key is the first line of the card, and the value is a list of string.

    *e.g.* ::

        qe_options = {
            '&control' : {
                'calculation' : "'scf'",
                'nstep' : 50,
                },
            'k_points automatic' : ['1 1 1 0 0 0'],
            }

    .. note ::

        If the type of value of namelist is string(character), the value also need contains with quotes.


    Parameters
    ----------
    filename : str
        Name of QE input file
    qe_options: dict
        A dictionary with input parameters for QE to generate QE input file.
    """

    def __init__(self, filename = None, qe_options = {}, **kwargs):
        self.filename = filename

        if self.filename is not None :
            self.qe_options = self.read_qe_input(self.filename, **kwargs)
            self.qe_options = self.update_options(options=qe_options, qe_options = self.qe_options)
        else :
            self.qe_options = qe_options

    def write_qe_input(self, filename, atoms = None, basefile = None, qe_options = {}, prog = 'pw', **kwargs):
        """Write the QE input file

        Parameters
        ----------
        filename : str
            The file name of output file
        atoms : ase.Atoms
            structure information with ase.Atoms
        basefile : str
            If set, will base on this file only update atomic information
        qe_options : dict
            Input parameters of QE
        """
        if basefile :
            options = qe_options
            qe_options = self.read_qe_input(basefile)
            qe_options = self.update_options(options=options, qe_options = qe_options, prog = prog)
        #
        if prog == 'pw' :
            if atoms is not None :
                qe_options = self.update_atoms(atoms, qe_options, prog=prog, **kwargs)
        if hasattr(filename, 'write'):
            fh = filename
        else :
            fh = open(filename, 'w')
        #
        options = QEOPTIONS.get(prog, {}).copy()
        options.update(qe_options)
        #
        for key, value in options.items() :
            if key.startswith('&'):
                fh.write(key.upper() + '\n')
                for k, v in value.items() :
                    fh.write('   ' + k + ' = ' + str(v) + '\n')
                fh.write('/\n\n')
            else :
                l = key.split()
                l[0] = l[0].upper()
                key = ' '.join(l)
                fh.write(key + '\n')
                for v in value :
                    fh.write(str(v) + '\n')
                fh.write('\n')

        if not hasattr(filename, 'write'): fh.close()

    @classmethod
    def update_atoms(cls, atoms, qe_options, prog = 'pw', **kwargs):
        """update atomic information

        Parameters
        ----------
        atoms : ase.Atoms
            atoms has all atomic information
        qe_options : dict
            Input parameters of QE
        """
        #
        if prog != 'pw' : return qe_options
        #
        if '&system' not in qe_options :
            qe_options['&system'] = {}
        qe_options['&system']['ibrav'] = 0
        qe_options['&system']['nat'] = len(atoms)
        ntyp = len(set(atoms.symbols))
        ntyp_old = len(qe_options.get('atomic_species', []))
        if ntyp > ntyp_old :
            raise ValueError("The number of 'ATOMIC_SPECIES' not fit the number of types of atoms.")
        qe_options['&system']['ntyp'] = ntyp_old
        # keys = [k.split()[0] for k in qe_options]
        keys = list(qe_options.keys())
        for key in keys :
            if key.startswith(('cell_parameters', 'atomic_positions')) :
                qe_options.pop(key, None)
        #
        key = 'cell_parameters angstrom'
        qe_options[key] = []
        for i in range(3):
            line = '{0[0]:.14f} {0[1]:.14f} {0[2]:.14f}'.format(atoms.cell[i])
            qe_options[key].append(line)
        #
        key = 'atomic_positions angstrom'
        qe_options[key] = []
        for s, p in zip(atoms.symbols, atoms.positions):
            line = '{0:4s} {1[0]:.14f} {1[1]:.14f} {1[2]:.14f}'.format(s, p)
            qe_options[key].append(line)
        #
        return qe_options

    @classmethod
    def update_options(cls, options = {}, qe_options = {}, **kwargs):
        """update options

        Use *options* to update *qe_options*

        Parameters
        ----------
        options : dict
            New options
        qe_options : dict
            QE options
        """

        options = options or {}
        qe_options = qe_options or {}
        keys = list(qe_options.keys())

        for k, v in options.items():
            k = k.lower()
            if k.startswith('&'):
                if k in qe_options :
                    qe_options[k].update(v)
                else :
                    qe_options[k] = v.copy()
            else :
                kc = k.split()[0]
                for item in keys :
                    if kc in item :
                        qe_options.pop(item)
                        break
                qe_options[k] = v
        # correct the qe_options
        for k, v in options.items():
            if 'cell_parameters' in k:
                qe_options['&system']['ibrav'] = 0
            elif 'atomic_positions' in k:
                qe_options['&system']['nat'] = len(v)
                ntyp = len(set([x.split()[0] for x in v]))
                ntyp_old = len(qe_options.get('atomic_species', []))
                if ntyp > ntyp_old :
                    raise ValueError("The number of 'ATOMIC_SPECIES' not fit the number of types of atoms.")
                qe_options['&system']['ntyp'] = ntyp_old
        return qe_options

    def read_qe_input(self, filename, **kwargs):
        """read the QE input file

        Parameters
        ----------
        filename : str
            file name of QE input file
        """
        options = {}
        with open(filename, 'r') as fr :
            fh = self.iter_lines(fr, sep='!', **kwargs)
            for line in fh :
                if line.startswith('&'):
                    options[line.split()[0].lower()] = self.read_namelist(fh, **kwargs)
                else :
                    fh.send(line)
                    break
            for line in fh :
                l = line.split()
                l[0] = l[0].lower()
                key = ' '.join(l)
                options[key] = self.read_card(fh, **kwargs)
        return options

    def read_namelist(self, fh, **kwargs):
        """read the namelist

        Parameters
        ----------
        fh : iter
            modified file handler
        """
        options = {}
        for line in fh :
            if line == '/' :
                break
            else :
                k, v = line.split('=')
                options[k.strip()] = v.strip(',').strip()
        else :
            raise ValueError("The namelist not closed with '/'.")
        return options

    def read_card(self, fh, **kwargs):
        """read the card

        The comments lines which start with '#' will be kept. Here assuming if all the characters of first word
        are alphabet letters and length is greater than 6, this line will be next card.

        Parameters
        ----------
        fh : iter
            modified file handler
        """
        l = []
        for line in fh :
            a = line.split()[0].replace('_', '')
            if a.isalpha() and len(a) > 6 : # assuming only card name has more than 6 characters
                fh.send(line)
                break
            else :
                l.append(line)
        return l

    def iter_lines(self, fh, sep='!', **kwargs):
        """return the non-comment part of non-empty line

        `sep` is the comment characters. The comment line or empty line will be skip.
        If the line is mixed, only the front non-comment part will return.

        Parameters
        ----------
        fh : object
            file handler
        sep : str, list or tuple
            The delimiter string, multiple delimiters can given by a list
        """
        if isinstance(sep, (tuple, list)):
            seps = sep[1:]
            sep = sep[0]
        else :
            seps = []
        for line in fh:
            for s in seps :
                line = line.replace(s,sep)
            line = line.split(sep)[0].strip()
            if len(line) > 0 :
                line = yield line
                if line is not None :
                    yield line
                    yield line


# All QE namelist from 7.1
QEOPTIONS={
    "all_currents" : OrderedDict.fromkeys(
        ["&energy_current"], {}),
    "bands" : OrderedDict.fromkeys(
        ["&bands"], {}),
    "bgw2pw" : OrderedDict.fromkeys(
        ["&input_bgw2pw"], {}),
    "cp" : OrderedDict.fromkeys(
        ["&control", "&system", "&electrons", "&ions", "&cell", "&press_ai", "&wannier", "atomic_species"], {}),
    "cppp" : OrderedDict.fromkeys(
        ["&inputpp"], {}),
    "davidson" : OrderedDict.fromkeys(
        ["&lr_input", "&lr_dav"], {}),
    "dos" : OrderedDict.fromkeys(
        ["&dos"], {}),
    "dynmat" : OrderedDict.fromkeys(
        ["&input"], {}),
    "eels" : OrderedDict.fromkeys(
        ["&lr_input", "&lr_control"], {}),
    "hp" : OrderedDict.fromkeys(
        ["&inputhp"], {}),
    "importexport_binary" : OrderedDict.fromkeys(
        ["&inputpp"], {}),
    "kcw" : OrderedDict.fromkeys(
        ["&control", "&wannier", "&screen", "&ham"], {}),
    "lanczos" : OrderedDict.fromkeys(
        ["&lr_input", "&lr_control", "&lr_post"], {}),
    "ld1" : OrderedDict.fromkeys(
        ["&input", "&inputp", "&test"], {}),
    "magnons" : OrderedDict.fromkeys(
        ["&lr_input", "&lr_control"], {}),
    "matdyn" : OrderedDict.fromkeys(
        ["&input"], {}),
    "molecularpdos" : OrderedDict.fromkeys(
        ["&inputmopdos"], {}),
    "neb" : OrderedDict.fromkeys(
        ["&path"], {}),
    "ph" : OrderedDict.fromkeys(
        ["&inputph"], {}),
    "postahc" : OrderedDict.fromkeys(
        ["&input"], {}),
    "pp" : OrderedDict.fromkeys(
        ["&inputpp", "&plot"], {}),
    "projwfc" : OrderedDict.fromkeys(
        ["&projwfc"], {}),
    "pw" : OrderedDict.fromkeys(
        ["&control", "&system", "&electrons", "&ions", "&cell", "&fcp", "&rism", "atomic_species"], {}),
    "pw2bgw" : OrderedDict.fromkeys(
        ["&input_pw2bgw"], {}),
    "pwcond" : OrderedDict.fromkeys(
        ["&inputcond"], {}),
    "q2r" : OrderedDict.fromkeys(
        ["&input"], {}),
    "spectrum" : OrderedDict.fromkeys(
        ["&lr_input"], {}),
    }



import qepy

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

fname = 'qe_in.in'
qepy.mp_global.mp_startup(my_world_comm=commf, start_images=True)

ndiag_ = 0
world_comm = qepy.mp_world.get_world_comm()
intra_pool_comm = qepy.mp_pools.get_intra_pool_comm()
intra_bgrp_comm = qepy.mp_bands.get_intra_bgrp_comm()
inter_bgrp_comm = qepy.mp_bands.get_inter_bgrp_comm()

qepy.mp_diag.mp_start_diag(ndiag_, world_comm, intra_pool_comm, do_distr_diag_inside_bgrp_ = True)
qepy.set_mpi_comm_4_solvers(intra_pool_comm, intra_bgrp_comm, inter_bgrp_comm)
qepy.environment.environment_start('PWSCF')
qepy.read_input.read_input_file('PW', fname)

exit_status = qepy.run_pwscf()

qepy.laxlib_free_ortho_group()
qepy.stop_run(exit_status)
qepy.do_stop(exit_status)



import qepy
from qepy.driver import Driver

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    comm = comm.py2f()
except Exception:
    comm = None

fname = 'qe_in.in'

driver = Driver(fname, comm)

for i in range(60):
    driver.diagonalize()
    driver.mix(mix_coef = 0.7)
    if driver.check_convergence(): break

energy = driver.get_energy()
forces = driver.get_forces()
stress = driver.get_stress()
driver.stop()

# print('energy', energy)
# print('forces', forces)
# print('stress', stress)



from qepy.driver import Driver
import numpy as np
import pathlib

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'

def test_scf():
    driver = Driver(inputfile, comm)
    driver.scf()
    converged = driver.check_convergence()
    #
    energy = driver.get_energy()
    forces = driver.get_forces()
    stress = driver.get_stress()
    if driver.is_root :
        print('converged :\n', converged)
        print('energy :\n', energy)
        print('forces :\n', forces)
        print('stress :\n', stress)
    assert converged
    assert np.isclose(energy, -552.93477389, atol = 1E-6)
    assert np.isclose(forces[0, 0], -0.00835135, atol = 1E-3)
    assert np.isclose(stress[1, 1], -0.00256059, atol = 1E-3)
    driver.stop()



from qepy.driver import Driver
import numpy as np
import pathlib

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

inputfile = 'qe_in.in'

def main():
    # Run scf
    driver = Driver(inputfile, comm)
    driver.scf()
    converged = driver.check_convergence()
    energy = driver.get_energy()
    driver.save()
    #
    energy = driver.get_energy()
    if driver.is_root :
        print('converged :\n', converged)
        print('energy :\n', energy)
    # Run TDDFT
    driver = Driver(inputfile, comm, task = 'optical')
    driver.scf()
    dipole = driver.get_dipole_tddft()
    if driver.is_root :
        print('dipople:\n', dipole)
    driver.stop()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8

import numpy as np
import qepy

try:
    from mpi4py import MPI
    comm=MPI.COMM_WORLD.py2f()
except Exception :
    comm=None

# qepy.qepy_mod.qepy_set_stdout('qepy.out')

qepy.qepy_pwscf('vcrelax.in',comm)
embed = qepy.qepy_common.embed_base()

nstep = qepy.control_flags.get_nstep()
#-------------------------------------------------------------------------------
lforce = qepy.force_mod.get_lforce()
lstres = qepy.force_mod.get_lstres()
sigma = qepy.force_mod.get_array_sigma()
lmovecell = qepy.cellmd.get_lmovecell()

fix_volume = qepy.cell_base.get_fix_volume()
fix_area = qepy.cell_base.get_fix_area()
treinit_gvecs = qepy.control_flags.get_treinit_gvecs()

ions_status = np.ones(1, dtype = 'int32')*3
#-------------------------------------------------------------------------------
for idone in range(0, nstep):
    if idone>0 : qepy.control_flags.set_ethr(1E-6)
    qepy.electrons()

    conv_ions = True

    if lforce : qepy.forces()
    if lstres : qepy.stress(sigma)

    lmd = qepy.control_flags.get_lmd()
    lbfgs = qepy.control_flags.get_lbfgs()

    if lmd or lbfgs :
        if fix_volume: qepy.impose_deviatoric_stress(sigma)
        if fix_area: qepy.impose_deviatoric_stress_2d(sigma)

        qepy.extrapolation.update_file()

        qepy.move_ions(idone, ions_status)

        conv_ions = (ions_status == 0) or (ions_status == 1 and treinit_gvecs)

        if idone < nstep and not conv_ions :
            qepy.punch('config-nowf')

        if qepy.funct.dft_is_hybrid() : qepy.funct.stop_exx()

    if conv_ions: break

    if lmd or lbfgs :
        if ions_status == 1 :
            qepy.control_flags.set_lbfgs(False)
            qepy.control_flags.set_lmd(False)
            print('Final scf calculation at the relaxed structure', flush=True)
            qepy.reset_gvectors()
        elif ions_status == 2 :
            qepy.reset_magn()
        else :
            if treinit_gvecs :
                if lmovecell : qepy.scale_h()
                qepy.reset_gvectors()
            else :
                qepy.extrapolation.update_pot()
                qepy.hinit1()

qepy.qepy_stop_run(0, what = 'no')



# import atexit

# def pwscf_finalise():
#     qepy_pwscf_finalise()


# atexit.register(pwscf_finalise)

import pkgutil
import operator
def qepy_clean_saved():
    mods = [name for _, name, _ in pkgutil.iter_modules(qepy.__path__)]
    for mod in mods :
        if hasattr(qepy, mod):
            mod += '._arrays'
            operator.attrgetter(mod)(qepy).clear()


qepy_clean_saved()
__author__ = "Pavanello Research Group"
__contact__ = "m.pavanello@rutgers.edu"
__version__ = "0.0.1"
__license__ = "GPL"
__date__ = "2023-02-14"

try:
    from importlib.metadata import version # python >= 3.8
except Exception :
    try:
        from importlib_metadata import version
    except Exception :
        pass

try:
    __version__ = version("qepy")
except Exception :
    pass



from qepy.driver import Driver
import numpy as np
import pathlib

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'

def test_0_scf():
    driver = Driver(inputfile, comm)
    driver.scf()
    converged = driver.check_convergence()
    energy = driver.get_energy()
    assert converged
    assert np.isclose(energy, -552.93477389, atol = 1E-6)
    driver.stop()

def test_1_read():
    driver = Driver(comm = comm, prefix = 'tmp', task = 'nscf')
    energy = driver.get_energy()
    if driver.is_root :
        print('energy :\n', energy)
    assert np.isclose(energy, -552.93477389, atol = 1E-6)
    driver.stop(what = 'no')

def test_2_read_pw():
    driver = Driver(inputfile, comm)
    driver.pwscf_restart()
    energy = driver.get_energy()
    if driver.is_root :
        print('energy :\n', energy)
    assert np.isclose(energy, -552.93477389, atol = 1E-6)
    driver.stop(what = 'no')



from qepy.calculator import QEpyCalculator
import numpy as np
import pathlib

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'

def test_md():
    import ase.io
    from ase import units
    from ase.md.verlet import VelocityVerlet

    atoms = QEpyCalculator(inputfile = inputfile, comm = comm, from_file = True).atoms

    dyn = VelocityVerlet(atoms, 2 * units.fs)
    # from ase.io.trajectory import Trajectory
    # traj = Trajectory("al_md_3.traj", "w", atoms)
    # dyn.attach(traj.write, interval=1)
    dyn.run(3)
    atoms.calc.stop()

    atoms_fin = ase.io.read(path / 'al_md_3.traj', index=-1)
    assert np.allclose(atoms.get_momenta(), atoms_fin.get_momenta(), atol=1.e-3)



import numpy as np
import qepy

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    comm = comm.py2f()
except Exception:
    comm = None

fname = 'qe_in.in'

qepy.qepy_pwscf(fname, comm)

embed = qepy.qepy_common.embed_base()
# embed.ldescf = True # add scf correction energy
embed.extene = 0.0
#
embed.iterative = True

for i in range(60):
    embed.mix_coef = -1.0
    qepy.qepy_electrons_scf(2, 0, embed)
    embed.mix_coef = 0.7
    qepy.qepy_electrons_scf(2, 0, embed)
    if qepy.control_flags.get_conv_elec() : break

# qepy.qepy_calc_energies(embed)
etotal = embed.etotal

#
nat = qepy.ions_base.get_nat()
forces = np.zeros((3, nat), order = 'F')
qepy.qepy_mod.qepy_set_extforces(embed, forces)
#
qepy.qepy_forces(embed = embed)
forces = qepy.force_mod.get_array_force().T

#
embed.extstress = 0.0
#
stress = np.ones((3, 3), order='F')
qepy.qepy_stress(stress, embed=embed)

qepy.punch('all')
qepy.qepy_stop_run(0, what = 'no')



#!/usr/bin/env python3
import qepy
from qepy.calculator import QEpyCalculator, units

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

import ase.io
from ase.io.trajectory import Trajectory

from ase.optimize import BFGS, LBFGS, FIRE
from ase.optimize.sciopt import SciPyFminBFGS, SciPyFminCG
from ase.constraints import StrainFilter, UnitCellFilter
from ase.spacegroup.symmetrize import FixSymmetry, check_symmetry

inputfile = 'vcrelax.in'

calc = QEpyCalculator(comm = comm, inputfile = inputfile, lmovecell = True)

atoms = ase.io.read(inputfile, format='espresso-in')

atoms.set_calculator(calc)
atoms.set_constraint(FixSymmetry(atoms))

# af = atoms 

pressure = 10 # GPa
af = UnitCellFilter(atoms, scalar_pressure = pressure*units['J']/10**21)

trajfile = 'opt.traj'
opt = LBFGS(af, trajectory = trajfile, memory = 8, use_line_search = False)

opt.run(fmax = 0.01)

if comm and comm.rank == 0 :
    traj = Trajectory(trajfile)
    ase.io.write('opt.vasp', traj[-1], direct = True, long_format=True, vasp5 = True)
    ase.io.write('opt.in', traj[-1], format = 'espresso-in')



import numpy as np
def conf2array(confs):
    import re
    pattern=re.compile(r'(\d*)([spdf])(\d*)')
    spdf={'s': 0, 'p': 1, 'd': 2, 'f': 3}
    nlocc = np.dtype([('symbol', np.unicode_, 3), ('econf', np.unicode_, 30), ('occ', np.int32, (3,))])
    data=[]
    for k,value in confs.items():
        if value is None or value.startswith('#') : continue
        vv = value.split('#')
        v=vv[0]
        m=pattern.match(v)
        n=m.group(1)
        l=spdf[m.group(2)]
        occ=m.group(3)
        if len(vv)>1 :
            econf=vv[1]
        else:
            econf=v
        if not occ : occ = 1
        data.append((k.capitalize(), econf.strip(), (n, l, occ)))
    occs = np.array(data, dtype=nlocc)
    return occs

def write_set_hubbard_n(occs, filename='set_hubbard_n.f90'):
    header="""!
! Copyright (C) 2001-2020 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!---------------------------------------------------------------------------
FUNCTION set_hubbard_n( psd ) RESULT( hubbard_n )
  !---------------------------------------------------------------------------
  !
  USE io_global, ONLY : stdout
  !
  IMPLICIT NONE
  !
  INTEGER                      :: hubbard_n
  CHARACTER(LEN=2), INTENT(IN) :: psd
  !
  !
  SELECT CASE( TRIM(ADJUSTL(psd)) )
"""
    ender="""
     CASE DEFAULT
        !
        hubbard_n = -1
        !
        WRITE( stdout, '(/,"psd = ",A,/)' ) psd
        !
        CALL errore( 'set_hubbard_n', 'pseudopotential not yet inserted', 1 )
        !
  END SELECT
  !
  RETURN
  !
END FUNCTION set_hubbard_n
"""
    fstr='     !qepy --> auto\n'
    index=np.lexsort((occs['occ'][:,2],occs['occ'][:,1],occs['occ'][:,0]))
    occs=occs[index]
    for i in range(1,9):
        index=occs['occ'][:,0]==i
        if occs[index].size > 0 :
            fstr+='     CASE('
            for nk, s in enumerate(occs[index]['symbol']):
                if nk>0 and nk % 10 == 0: fstr += '&\n          '
                fstr += f"'{s}', "
            fstr = fstr[:-2]+')\n'
            fstr += f'        hubbard_n = {i}\n'
    fstr+='     !qepy <-- auto\n'
#             print(i, fstr)
    fstr=header+fstr+ender
    with open(filename, 'w') as fh:
        fh.write(fstr)

def write_set_hubbard_l(occs, filename='set_hubbard_l.f90'):
    header="""!
! Copyright (C) 2001-2018 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!---------------------------------------------------------------------------
FUNCTION set_hubbard_l( psd ) RESULT( hubbard_l )
  !---------------------------------------------------------------------------
  !
  USE io_global, ONLY : stdout
  !
  IMPLICIT NONE
  !
  INTEGER                      :: hubbard_l
  CHARACTER(LEN=2), INTENT(IN) :: psd
  !
  !
  SELECT CASE( TRIM(ADJUSTL(psd)) )
"""
    ender="""
     CASE DEFAULT
        !
        hubbard_l = -1
        !
        WRITE( stdout, '(/,"psd = ",A,/)' ) psd
        !
        CALL errore( 'set_hubbard_l', 'pseudopotential not yet inserted', 1 )
        !
  END SELECT
  !
  RETURN
  !
END FUNCTION set_hubbard_l
"""
    fstr='     !qepy --> auto\n'
    index=np.lexsort((occs['occ'][:,2],occs['occ'][:,1],occs['occ'][:,0]))
    occs=occs[index]
    for i in range(1,9):
        index=occs['occ'][:,0]==i
        occs2=occs[index]
        if occs2.size > 0 :
            fstr+=f'     ! hubbard_n -> {i}\n'
            for j in range(0,4):
                index=occs2['occ'][:,1]==j
                if occs2[index].size > 0:
                    fstr+='     CASE('
                    for nk, s in enumerate(occs2[index]['symbol']):
                        fstr += f"'{s}', "
                    fstr = fstr[:-2]+')\n'
                    fstr += f'        hubbard_l = {j}\n'
    fstr+='     !qepy <-- auto\n'
    fstr=header+fstr+ender
#     print(fstr)
    with open(filename, 'w') as fh:
        fh.write(fstr)

def write_tabd(occs, filename='tabd.f90'):
    header="""!
! Copyright (C) 2001-2020 Quantum ESPRESSO group
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!-----------------------------------------------------------------------
FUNCTION hubbard_occ( psd )
  !-----------------------------------------------------------------------
  !! This routine is a table (far from being complete) for the total number
  !! of localized electrons in transition metals or rare earths (PPs usually
  !! are built on non physical configurations).
  !
  USE kinds, ONLY: DP
  !
  IMPLICIT NONE
  !
  CHARACTER(LEN=2), INTENT(IN) :: psd
  REAL(DP)                     :: hubbard_occ
  !
  SELECT CASE( TRIM(ADJUSTL(psd)) )
"""
    ender="""
     !
     !
     ! NOT INSERTED
     !
     CASE DEFAULT
        hubbard_occ = 0.d0
        call errore ('hubbard_occ', 'pseudopotential not yet inserted', 1)
     !
  END SELECT
  !
  RETURN
  !
END FUNCTION hubbard_occ
"""
    fstr='     !qepy --> auto\n'
    index=np.lexsort((occs['occ'][:,2],occs['occ'][:,1],occs['occ'][:,0]))
    occs=occs[index]
    for item in occs:
        fstr+='     CASE('
        s=item['symbol']
        fstr += f"'{s}')"
        fstr +=f'     ! {item["symbol"]}->{item["econf"]}\n'
        fstr += f'        hubbard_occ = {str(item["occ"][2])}d0\n'
    fstr+='     !qepy <-- auto\n'
    fstr=header+fstr+ender
#     print(fstr)
    with open(filename, 'w') as fh:
        fh.write(fstr)

def get_ldau_occs(filename):
    from configparser import ConfigParser
    config = ConfigParser()
    config.read(filename)
    confs=config['DEFAULT']
    occs=conf2array(confs)
    return occs

def ini2files(filename = 'qepy_econf.ini'):
    occs = get_ldau_occs(filename)
    write_set_hubbard_n(occs)
    write_set_hubbard_l(occs)
    write_tabd(occs)


if __name__ == '__main__':
    import sys
    filename = sys.argv[1] if len(sys.argv)>1 else 'qepy_econf.ini'
    ini2files(filename)



import numpy as np
import qepy
import unittest
import pathlib
import shutil
try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'


class Test(unittest.TestCase):
    def test_scf(self):
        qepy.qepy_pwscf(inputfile, commf)
        embed = qepy.qepy_common.embed_base()
        # embed.ldescf = True # add scf correction energy
        embed.iterative = True
        for i in range(60):
            embed.mix_coef = -1.0
            qepy.qepy_electrons_scf(2, 0, embed)
            embed.mix_coef = 0.7
            qepy.qepy_electrons_scf(2, 0, embed)
            if qepy.control_flags.get_conv_elec() : break

        conv_flag = bool(qepy.control_flags.get_conv_elec())
        self.assertTrue(conv_flag)

        etotal = embed.etotal
        self.assertTrue(np.isclose(etotal, -552.93477389, atol = 1E-6))

        qepy.qepy_stop_run(0, what = 'no')

    def tearDown(self):
        if comm and comm.rank == 0 :
            path = pathlib.Path('.')
            for f in path.glob('al.*'):
                if f.is_file():
                    f.unlink()
                else :
                    shutil.rmtree(f)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import qepy
import unittest
import pathlib
import shutil

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'


class Test(unittest.TestCase):
    def test_0_scf(self):
        qepy.qepy_pwscf(inputfile, commf)
        qepy.electrons()

        conv_flag = bool(qepy.control_flags.get_conv_elec())
        self.assertTrue(conv_flag)

        etotal = qepy.ener.get_etot()
        self.assertTrue(np.isclose(etotal, -552.93477389, atol = 1E-6))

        qepy.punch('all')

    def test_1_tddft_continue(self):
        embed = qepy.qepy_common.embed_base()
        qepy.qepy_tddft_readin(inputfile)
        qepy.qepy_tddft_main_setup(embed)
        qepy.qepy_molecule_optical_absorption(embed)
        qepy.qepy_stop_run(0, print_flag=0, what='no', finalize=False)
        qepy.qepy_stop_tddft(0)

    def test_2_tddft_iterative(self):
        qepy.qepy_tddft_main_initial(inputfile, commf)
        qepy.read_file()
        embed = qepy.qepy_common.embed_base()
        embed.tddft.iterative = True
        qepy.qepy_tddft_main_setup(embed)

        for i in range(5):
            qepy.qepy_molecule_optical_absorption(embed)
        dip = qepy.qepy_tddft_common.get_array_dipole().copy()
        embed.tddft.finish = True
        qepy.qepy_molecule_optical_absorption(embed)
        qepy.qepy_stop_tddft(0)

        assert(abs(dip[0, 0] - 0.54355)<1E-3)

    def test_9_clean(self):
        if comm and comm.rank == 0 :
            path = pathlib.Path('.')
            for f in path.glob('al.*'):
                if f.is_file():
                    f.unlink()
                else :
                    shutil.rmtree(f)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import qepy
import unittest
import pathlib
import shutil

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

path = pathlib.Path(__file__).resolve().parent / 'DATA'
inputfile = path / 'qe_in.in'


class Test(unittest.TestCase):
    def test_0_scf(self):
        qepy.qepy_pwscf(inputfile, commf)
        qepy.electrons()

        conv_flag = bool(qepy.control_flags.get_conv_elec())
        self.assertTrue(conv_flag)

        etotal = qepy.ener.get_etot()
        self.assertTrue(np.isclose(etotal, -552.93477389, atol = 1E-6))

        qepy.qepy_stop_run(0, what = 'all')

    def test_1_read(self):
        inputobj = qepy.qepy_common.input_base()
        inputobj.prefix = 'tmp'
        if commf : inputobj.my_world_comm = commf

        qepy.qepy_initial(inputobj)

        qepy.qepy_read_file()

        embed = qepy.qepy_common.embed_base()
        qepy.qepy_calc_energies(embed)
        self.assertTrue(np.isclose(embed.etotal, -552.93477389, atol = 1E-6))
        qepy.qepy_stop_run(0, what = 'no')

    def test_2_read_pw(self):
        qepy.qepy_pwscf(inputfile, commf)
        embed = qepy.qepy_common.embed_base()

        qepy.qepy_pw_restart_new.qepy_read_xml_file(alloc=False)
        if qepy.basis.get_starting_pot().strip() != 'file' :
            qepy.qepy_potinit(starting = 'file')
        if qepy.basis.get_starting_wfc().strip() != 'file' :
            qepy.qepy_wfcinit(starting = 'file')

        qepy.qepy_calc_energies(embed)
        self.assertTrue(np.isclose(embed.etotal, -552.93477389, atol = 1E-6))
        qepy.qepy_stop_run(0, what = 'no')

    @classmethod
    def tearDownClass(cls):
        if comm and comm.rank == 0 :
            path = pathlib.Path('.')
            for f in path.glob('al.*'):
                if f.is_file():
                    f.unlink()
                else :
                    shutil.rmtree(f)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import tempfile
import qepy
from qepy.core import Logger, env
from qepy.io import QEInput
from qepy import constants

class Driver(metaclass = Logger) :
    """
    The driver of QEpy.

    Parameters
    ----------
    inputfile : str
        Name of QE input file
    comm : object
        Parallel communicator
    ldescf : bool
        If True, also print the scf correction term in the iteratively mode
    iterative : bool
        Iteratively run the scf or tddft
    task : str
        Task of the driver :

          - 'scf' : scf task
          - 'optical' : Optical absorption spectrum (TDDFT)
          - 'nscf' : read file from scf

    embed : object (Fortran)
        embed object for QE
    prefix : str
        prefix of QE input/output
    outdir : str
        The output directory of QE
    logfile : str, file or bool
        The screen output of QE. If it is a file descriptor, it should open with 'w+'.

          - None : Show on the screen.
          - str  : Save to the given file.
          - True : Save to the temporary file, see also :func:`qepy.driver.Driver.get_output`.

    outdir : str
        The output directory of QE
    qe_options : dict
        A dictionary with input parameters for QE to generate QE input file.
    prog : str
        The name of QE program, default is `pw` which is pw.x in QE.
    progress : bool
        If True, will continue run the QE without clean the workspace, most times used for TDDFT after scf.
    kwargs : dict
        Other options


    .. note::

         - The bool `gather` parameter in functions means the data is gathered in root or distributed in all cpus.
         - Also contains some interface of ASE calculator, but the units are different.

             + positions : Bohr
             + lattice : Bohr
             + energy : Ry
             + forces : Ry/Bohr
             + stress : Ry/Bohr**3
             + potential : Ry
             + density : 1.0/Bohr**3

    """

    def __init__(self, inputfile = None, comm = None, ldescf = False, iterative = False,
             task = 'scf', embed = None, prefix = None, outdir = None, logfile = None,
             qe_options = None, prog = 'pw', progress = False, **kwargs):
        if embed is None :
            embed = qepy.qepy_common.embed_base()
        self.task = task
        self.embed = embed
        self.comm = comm
        self.inputfile = inputfile
        self.iterative = iterative
        self.prefix = prefix
        self.outdir = outdir
        self.logfile = logfile
        self.qe_options = qe_options
        self.prog = prog
        self.progress = progress
        #
        self.embed.ldescf = ldescf
        #
        self.comm = comm
        self.qepy = qepy
        self.qeinput = QEInput()
        #
        self._init_log()
        #
        self.driver_initialize()

    def _init_log(self):
        """_initialize the QE output."""
        self.fileobj_interact = False
        if self.logfile is not None :
            # qepy.qepy_mod.qepy_set_stdout(self.logfile)
            if isinstance(self.logfile, bool) and self.logfile :
                self.fileobj_interact = True
                self.fileobj = tempfile.NamedTemporaryFile('w+')
            elif hasattr(self.logfile, 'write'):
                self.fileobj = self.logfile
            else :
                self.fileobj = open(self.logfile, 'w+')
        else :
            self.fileobj = None
        return self.fileobj

    @property
    def comm(self):
        return self._comm

    @comm.setter
    def comm(self, value):
        self._comm = value
        if self.comm is not None and hasattr(self.comm, 'py2f') :
            self.commf = self.comm.py2f()
        else :
            self.commf = self.comm

    def restart(self, prog=None, **kwargs):
        prog = prog or self.prog
        self._init_log()
        self.driver_initialize()

    def driver_initialize(self, **kwargs):
        """ Initialize the driver

        Parameters
        ----------
        inputfile : str
            Name of QE input file
        commf : object
            Parallel communicator (Fortran)
        iterative : bool
            Iteratively run the scf or tddft
        task : str
            Task of the driver :

              - 'scf' : scf task
              - 'optical' : Optical absorption spectrum (TDDFT)
              - 'nscf' : read file from scf

        qe_options: dict
            A dictionary with input parameters for QE to generate QE input file.

        """
        # stop the last driver and save new driver
        if not self.progress :
            if hasattr(env['DRIVER'], 'stop'): env['DRIVER'].stop()
            env['DRIVER'] = self
        #
        inputfile=self.inputfile
        commf=self.commf
        task=self.task
        qe_options = self.qe_options
        prog = self.prog
        #
        if qe_options :
            inputfile, basefile = 'input_tmp.in', inputfile
            self.qeinput.write_qe_input(inputfile, basefile=basefile, qe_options=qe_options, prog=prog)
        #
        if task == 'optical' :
            self.tddft_initialize(inputfile=inputfile, commf = commf, embed = self.embed, **kwargs)
        elif task == 'nscf' :
            inputobj = qepy.qepy_common.input_base()
            if self.prefix : inputobj.prefix = self.prefix
            if self.outdir : inputobj.tmp_dir = self.outdir
            if commf : inputobj.my_world_comm = commf
            qepy.qepy_initial(inputobj)
            qepy.qepy_read_file()
        else :
            qepy.qepy_pwscf(inputfile, commf, embed = self.embed)
            self.embed.iterative = self.iterative
            if self.embed.iterative :
                qepy.control_flags.set_niter(1)
        #
        self.density = np.zeros((1, 1))
        self.iter = 0

    def tddft_initialize(self, inputfile = None, commf = None, embed = None, **kwargs):
        """ Initialize the tddft

        Parameters
        ----------
        inputfile : str
            Name of QE input file, which also contains `&inputtddft` section.
        commf : object
            Parallel communicator (Fortran)
        embed : object (Fortran)
            embed object for QE
        """
        if inputfile is None : inputfile = self.inputfile
        if commf is None : commf = self.commf
        if embed is None : embed = self.embed
        #
        if self.progress :
            qepy.wvfct.get_array_g2kin()
            qepy.qepy_tddft_readin(inputfile)
        else :
            qepy.qepy_tddft_main_initial(inputfile, commf)
            qepy.read_file()
        qepy.qepy_tddft_main_setup(embed)
        self.embed.tddft.iterative = self.iterative

    def diagonalize(self, print_level = 2, **kwargs):
        """Diagonalize the hamiltonian

        Parameters
        ----------
        print_level :
            The level of output of QE
        """
        self.iter += 1
        if self.task == 'optical' :
            qepy.qepy_molecule_optical_absorption(self.embed)
        else :
            self.embed.mix_coef = -1.0
            qepy.qepy_electrons_scf(print_level, 0, self.embed)

    def mix(self, mix_coef = 0.7, print_level = 2):
        """Mixing the density

        Parameters
        ----------
        print_level :
            The level of output of QE
        """
        if self.task == 'optical' : return
        self.embed.mix_coef = mix_coef
        qepy.qepy_electrons_scf(print_level, 0, self.embed)

    def check_convergence(self, **kwargs):
        """Check the convergence of the SCF"""
        converged = bool(qepy.control_flags.get_conv_elec())
        if converged and not self.embed.initial : self.end_scf()
        return converged

    def get_scf_error(self, **kwargs):
        """Return the error of the scf"""
        if self.embed.iterative :
            return self.embed.dnorm
        else :
            return qepy.control_flags.get_scf_error()

    def get_scf_steps(self, **kwargs):
        """Return the number of steps of scf"""
        if self.embed.iterative :
            return self.iter
        else :
            return qepy.control_flags.get_n_scf_steps()

    def scf(self, print_level = 2, maxiter = None, original = False, **kwargs):
        """Run the scf/tddft until converged or maximum number of iterations"""
        if maxiter is not None and not self.embed.iterative :
            qepy.control_flags.set_niter(maxiter)
        if self.task == 'optical' :
            qepy.qepy_molecule_optical_absorption(self.embed)
        elif not self.embed.iterative and self.embed.exttype < 2 :
            # Use electrons to support hybrid xc functional
            return self.electrons(original=original)
        else :
            qepy.qepy_electrons_scf(print_level, 0, self.embed)
        return self.embed.etotal

    def non_scf(self, **kwargs):
        # fix some saved variables from last scf calculations
        qepy.control_flags.set_lscf(0)
        qepy.control_flags.set_lbfgs(0)
        qepy.control_flags.set_lmd(0)
        qepy.control_flags.set_lwf(0)
        #
        qepy.non_scf()
        return qepy.ener.get_etot()

    def electrons(self, original = False, **kwargs):
        if original :
            qepy.electrons()
        else :
            qepy.qepy_electrons(self.embed)
        return qepy.ener.get_etot()

    def end_scf(self, **kwargs):
        """End the scf and clean the scf workspace. Only need run it in iterative mode"""
        if self.embed.iterative :
            self.embed.finish = True
            qepy.qepy_electrons_scf(0, 0, self.embed)

    def stop(self, exit_status = 0, what = 'all', print_flag = 0, **kwargs):
        """stop.

        Parameters
        ----------
        exit_status : int
            0 : QE will remove it temporary files
        print_flag : int
            0 : Not print time informations
        what : str
             see :func:`qepy.driver.Driver.save`.
        """
        #
        env['DRIVER'] = None
        #
        if self.task == 'optical' :
            self.tddft_stop(exit_status, print_flag = print_flag, what = what, **kwargs)
        else :
            qepy.qepy_stop_run(exit_status, print_flag = print_flag, what = what, finalize = False)

        if hasattr(self.fileobj, 'close'): self.fileobj.close()

    def tddft_restart(self, istep=None, **kwargs):
        """Restart the tddft from previous interrupted run.

        Parameters
        ----------
        istep : int
            Start number of steps, just for output.
        """
        qepy.qepy_tddft_mod.qepy_cetddft_wfc2rho()
        if istep is not None :
            self.embed.tddft.istep = istep

    def tddft_stop(self, exit_status = 0, what = 'no', print_flag = 0, **kwargs):
        if self.embed.tddft.iterative :
            self.embed.tddft.finish = True
            qepy.qepy_molecule_optical_absorption(self.embed)
        #! Do not save the PW files, otherwise the initial wfcs will be overwritten.
        qepy.qepy_stop_run(exit_status, print_flag = print_flag, what = 'no', finalize = False)
        qepy.qepy_stop_tddft(exit_status)

    def save(self, what = 'all', **kwargs):
        """
        Save the QE data to the disk

        Parameters
        ----------
        what : str

          * what = 'all' : write xml data file, charge density, wavefunctions
                           (for final data);
          * what = 'config' : write xml data file and charge density; also,
                              for nks=1, wavefunctions in plain binary format
                              (see why in comments below). For intermediate
                              or incomplete results;
          * what = 'config-nowf' : write xml data file and charge density only;
                           (save density);
          * what = 'config-init' : write xml data file only excluding final results
                                   (for dry run, can be called at early stages).

          see PW/src/punch.f90
        """
        qepy.punch(what)

    def update_run_options(self, qe_options = {}, **kwargs):
        pass

    def get_energy(self, **kwargs):
        """Return the total energy."""
        if abs(qepy.ener.get_etot()) > 1E-16 :
            energy = qepy.ener.get_etot()
        elif abs(self.embed.etotal) > 1E-16 :
            energy = self.embed.etotal
        else :
            energy = self.calc_energy(**kwargs)
        return energy

    def calc_energy(self, **kwargs):
        """Calculate the energy with the pw2casino of QE."""
        qepy.qepy_calc_energies(self.embed)
        return self.embed.etotal

    def update_ions(self, positions = None, lattice = None, update = 0, **kwargs):
        """update the ions of QE

        Parameters
        ----------
        positions : np.ndarray (n, 3)
            positions of ions
        lattice : 3x3 matrix
            lattice of ions
        update : int
             - 0 : update everything (default)
             - 1 : only update atomic information, not update potential

        """
        positions = positions.T
        if lattice is not None :
            lattice = lattice.T
            if not qepy.cellmd.get_lmovecell():
                raise ValueError(" Lattice update only works for variable-cell simulations.\n Please restart the QEpy with calculation= 'vc-relax' or 'vc-md'")
            qepy.qepy_api.qepy_update_ions(self.embed, positions, update, lattice)
        else :
            qepy.qepy_api.qepy_update_ions(self.embed, positions, update)

    def pwscf_restart(self, oldxml=False):
        """Read PW ouput/restart files.

        Parameters
        ----------
        oldxml : bool
             - True : read the old format xml file (qe<6.4)
             - False : read the new format xml file (qe>6.3)
        """
        if oldxml :
            if not hasattr(qepy, 'oldxml_pw_restart') :
                raise AttributeError("Please reinstall the QEpy with 'oldxml=yes'.")
            qepy.oldxml_pw_restart.pw_readfile('header')
            qepy.oldxml_pw_restart.pw_readfile('reset')
            qepy.oldxml_pw_restart.pw_readfile('dim')
            qepy.oldxml_pw_restart.pw_readfile('bs')
            if qepy.basis.get_starting_pot().strip() != 'file' :
                qepy.oldxml_potinit(starting = 'file')
            if qepy.basis.get_starting_wfc().strip() != 'file' :
                qepy.oldxml_wfcinit(starting = 'file')
        else :
            qepy.qepy_pw_restart_new.qepy_read_xml_file(alloc=False)
            if qepy.basis.get_starting_pot().strip() != 'file' :
                qepy.qepy_potinit(starting = 'file')
            if qepy.basis.get_starting_wfc().strip() != 'file' :
                qepy.qepy_wfcinit(starting = 'file')

    @property
    def is_root(self):
        """Whether it is the root.

        If the comm is set, root is rank == 0.
        Otherwise root is ionode of QE.
        """
        if hasattr(self.comm, 'rank'):
            return self.comm.rank == 0
        else :
            return qepy.io_global.get_ionode()

    @staticmethod
    def get_ions_lattice():
        """Return the lattice of ions."""
        alat = qepy.cell_base.get_alat()
        lattice = qepy.cell_base.get_array_at() * alat
        return lattice.T

    @staticmethod
    def get_ions_positions():
        """Return the cartesian positions of ions."""
        alat = qepy.cell_base.get_alat()
        pos = qepy.ions_base.get_array_tau().T * alat
        return pos

    @staticmethod
    def get_ions_symbols():
        """Return the symbols of ions."""
        ityp = qepy.ions_base.get_array_ityp() - 1
        nat = qepy.ions_base.get_nat()
        ntyp = qepy.ions_base.get_nsp()
        label = qepy.ions_base.get_array_atm().T.view('S3')[:,0].astype('U3')[:ntyp]
        label = [x.strip() for x in label]
        symbols = []
        for i in range(nat):
            symbols.append(label[ityp[i]])
        return symbols

    def get_density(self, gather = True):
        """Return density array in real space."""
        nr = self.get_number_of_grid_points(gather = gather)
        nspin = qepy.lsda_mod.get_nspin()
        if gather :
            if self.is_root :
                if np.prod(nr) != self.density.shape[0] or nspin != self.density.shape[1] :
                    self.density = np.empty((np.prod(nr), nspin), order = 'F')
            else :
                self.density = np.empty((1, nspin), order = 'F')
        else :
            if np.prod(nr) != self.density.shape[0] or nspin != self.density.shape[1] :
                self.density = np.empty((np.prod(nr), nspin), order = 'F')
        qepy.qepy_mod.qepy_get_rho(self.density, gather = gather)
        return self.density

    def get_wave_function(self, band=None, kpt=0):
        """Return wave-function array in real space."""
        qepy.qepy_mod.qepy_get_evc(kpt + 1)
        nrs = np.zeros(3, dtype = 'int32')
        qepy.qepy_mod.qepy_get_grid_smooth(nrs)
        if self.is_root :
            wf = np.empty(np.prod(nrs), order = 'F', dtype = np.complex128)
        else :
            wf = np.empty(1, order = 'F', dtype = np.complex128)
        if band is None :
            band = np.arange(self.get_number_of_bands())
        else :
            band = np.asarray(band)
            if band.ndim == 0 : band = [band]
        wfs = []
        for ibnd in band :
            qepy.qepy_mod.qepy_get_wf(kpt + 1, ibnd + 1, wf)
            wfs.append(wf.copy())
        return wfs

    def get_number_of_k_points(self):
        """Return the number of kpoints."""
        return qepy.klist.get_nks()

    def get_dipole_tddft(self):
        """Return the total dipole of tddft task."""
        # dipole = qepy.qepy_tddft_common.get_array_dipole().copy()
        dipole = self.embed.tddft.dipole
        return dipole

    def set_external_potential(self, potential, exttype = None, gather = True, extene  = None, **kwargs):
        """Set the external potential.

        Parameters
        ----------
        potential : (nnr, nspin)
            The external potential
        exttype : int
            The type of external potential

                 ==== ============================== ===
                 type potential                      bin
                 ==== ============================== ===
                  0   external                       000
                  1   pseudo                         001
                  2   hartree                        010
                  3   pseudo + hartree               011
                  4   xc                             100
                  5   pseudo + xc                    101
                  6   hartree + xc                   110
                  7   pseudo + hartree + xc          111
                 ==== ============================== ===

        """
        if exttype is not None :
            self.embed.exttype = exttype
        if extene is not None :
            self.embed.extene = extene
        else :
            self.embed.extene = 0.0
        #
        if potential is None : potential = np.zeros((1, 1))
        if potential.ndim != 2 : raise ValueError("The array should be 2-d.")
        #
        qepy.qepy_mod.qepy_set_extpot(self.embed, potential, gather = gather)

    def get_output(self):
        """Return the output of QE.

        It depends on the `logfile` of the initialization of the driver :

          - None : return None.
          - str  : return all outputs.
          - True : It will return the output from last time.
        """
        if self.fileobj is not None :
            if self.fileobj_interact :
                self.fileobj.seek(0)
                lines = self.fileobj.readlines()
                self.fileobj.close()
                self.fileobj = tempfile.NamedTemporaryFile('w+')
                return lines
            else :
                self.fileobj.seek(0)
                return self.fileobj.readlines()
        else :
            return None

    #ASE Calculator
    def get_potential_energy(self, **kwargs):
        """Return the potential energy."""
        return self.get_energy()

    def get_forces(self, icalc = 0, **kwargs):
        """Return the total forces. (n, 3)

        Parameters
        ----------
        icalc : int

            ===== ============================= ===
            icalc without                       bin
            ===== ============================= ===
              0   all                           000
              1   no ewald                      001
              2   no local                      010
              3   no ewald + local              011
              4   no nlcc                       100
              5   no ewald + nlcc               101
              6   no local + nlcc               110
              7   no ewald + local + nlcc       111
            ===== ============================= ===
        """
        qepy.qepy_forces(icalc, self.embed)
        forces = qepy.force_mod.get_array_force().T
        return forces

    def get_stress(self, **kwargs):
        """Return the stress (3, 3)."""
        stress = np.zeros((3, 3), order='F')
        qepy.stress(stress)
        # qepy.qepy_stress(stress, 0, self.embed)
        return stress

    #ASE DFTCalculator
    def get_number_of_bands(self):
        """Return the number of bands."""
        return qepy.wvfct.get_nbnd()

    def get_xc_functional(self):
        """Return the XC-functional identifier.

        'LDA', 'PBE', ..."""
        return qepy.funct.get_dft_short().decode("utf-8")

    def get_bz_k_points(self):
        """Return all the k-points in the 1. Brillouin zone.

        The coordinates are relative to reciprocal latice vectors."""
        return qepy.klist.get_array_xk()

    def get_number_of_spins(self):
        """Return the number of spins in the calculation.

        Spin-paired calculations: 1, spin-polarized calculation: 2."""
        return qepy.lsda_mod.get_nspin()

    def get_spin_polarized(self):
        """Is it a spin-polarized calculation?"""
        return bool(qepy.lsda_mod.get_lsda())

    def get_ibz_k_points(self):
        """Return k-points in the irreducible part of the Brillouin zone.

        The coordinates are relative to reciprocal latice vectors."""
        xk = qepy.klist.get_array_xk()[:, :self.get_number_of_k_points()].T
        return xk @ qepy.cell_base.get_array_at()

    def get_k_point_weights(self):
        """Weights of the k-points.

        The sum of all weights is one."""
        return qepy.klist.get_array_wk()[:self.get_number_of_k_points()]

    def get_pseudo_density(self, spin=None, pad=True, gather = True):
        """Return pseudo-density array.

        If *spin* is not given, then the total density is returned.
        Otherwise, the spin up or down density is returned (spin=0 or
        1)."""
        density = self.get_density(gather = gather)
        if spin is None :
            return density.sum(axis=1)
        else :
            return density[:, spin]

    def get_effective_potential(self, spin=0, pad=True):
        """Return pseudo-effective-potential array."""
        return qepy.scf.get_array_vrs()

    def get_pseudo_wave_function(self, band=None, kpt=0, spin=0, broadcast=True,
                                 pad=True):
        """Return pseudo-wave-function array."""
        qepy.qepy_mod.qepy_get_evc(kpt + 1)
        evc = qepy.wavefunctions.get_array_evc()
        if band is None :
            return evc
        else :
            return evc[:, band]

    def get_eigenvalues(self, kpt=0, spin=0):
        """Return eigenvalue array."""
        return qepy.wvfct.get_array_et()[:, kpt]

    def get_occupation_numbers(self, kpt=0, spin=0):
        """Return occupation number array."""
        return qepy.wvfct.get_array_wg()[:, kpt]

    def get_fermi_level(self):
        """Return the Fermi level."""
        return qepy.ener.get_ef()

    # def initial_wannier(self, initialwannier, kpointgrid, fixedstates,
                        # edf, spin, nbands):
        # """Initial guess for the shape of wannier functions.

        # Use initial guess for wannier orbitals to determine rotation
        # matrices U and C.
        # """
        # raise NotImplementedError

    # def get_wannier_localization_matrix(self, nbands, dirG, kpoint,
                                        # nextkpoint, G_I, spin):
        # """Calculate integrals for maximally localized Wannier functions."""
        # raise NotImplementedError

    def get_magnetic_moment(self, **kwargs):
        """Return the total magnetic moment."""
        return qepy.lsda_mod.get_magtot()

    def get_number_of_grid_points(self, gather = True):
        """Return the shape of arrays."""
        nr = np.zeros(3, dtype = 'int32')
        qepy.qepy_mod.qepy_get_grid(nr, gather)
        return nr
    #ASE DFTCalculator END

    def get_volume(self):
        """Return the volume."""
        return qepy.cell_base.get_omega()

    def get_ecutrho(self):
        """Return the cutoff for density."""
        return qepy.gvect.get_ecutrho()

    def data2field(self, data, cell = None, grid = None):
        """QE data to dftpy DirectField, please call it in serial."""
        from dftpy.field import DirectField
        from dftpy.grid import DirectGrid
        #
        if cell is None : cell = self.get_ions_lattice()
        if grid is None : grid = DirectGrid(lattice=cell, nr=self.get_number_of_grid_points(), ecut=self.get_ecutrho())
        field = DirectField(grid=grid, data=data.ravel(order='C'), order='F', rank=self.get_number_of_spins())
        return field

    def field2data(self, field, data = None):
        """dftpy DirectField to QE data, please call it in serial."""
        nspin = self.get_number_of_spins()
        if data is None :
            nnrR = np.prod(self.get_number_of_grid_points())
            data = np.empty((nnrR, nspin))
        #
        ns = field.shape[0] if field.ndim == 4 else 1
        if ns == 1 : field = [field]
        for i in range(nspin):
            data[:, i] = field[i].ravel(order = 'F')
        return data

    @classmethod
    def get_ase_atoms(cls):
        """Return the atom.Atoms from QE."""
        from ase.atoms import Atoms
        units_Bohr = constants.BOHR_RADIUS_SI * 1E10
        #
        symbols = cls.get_ions_symbols()
        positions = cls.get_ions_positions() * units_Bohr
        lattice = cls.get_ions_lattice() * units_Bohr
        atoms = Atoms(symbols = symbols, positions = positions, cell = lattice)
        return atoms

    @classmethod
    def get_dftpy_ions(cls):
        """Return the dftpy.Ions from QE."""
        from dftpy.ions import Ions
        atoms = cls.get_ase_atoms()
        return Ions.from_ase(atoms)

    def set_density(self, density, gather = True, **kwargs):
        """Set density array in real space."""
        #
        if density is None : density = np.zeros((1, 1))
        if density.ndim != 2 : raise ValueError("The array should be 2-d.")
        #
        qepy.qepy_mod.qepy_set_rho(density, gather = gather)



import qepy
try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    comm = comm.py2f()
except Exception:
    comm = None

oldxml = False # oldxml version QE

inputobj = qepy.qepy_common.input_base()
#-----------------------------------------------------------------------
inputobj.prefix = 'tmp'
# inputobj.tmp_dir = './tmp.wfx/'
#-----------------------------------------------------------------------
if comm : inputobj.my_world_comm = comm
qepy.qepy_initial(inputobj)

if oldxml :
    qepy.oldxml_read_file()
else :
    qepy.qepy_read_file()

embed = qepy.qepy_common.embed_base()
energy = qepy.qepy_calc_energies(embed)
qepy.qepy_stop_run(0, what = 'no')
# Newer version QE (>6.3) (XML file name is 'data-file-schema.xml')
# Olderversion QE (XML file name is 'data-file.xml')



from qepy.driver import Driver
import numpy as np
import pathlib

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

inputfile = 'qe_in.in'

def main():
    # Run TDDFT iterative
    driver = Driver(inputfile, comm, task = 'optical', iterative = True)
    for i in range(5):
        driver.diagonalize()
    dipole = driver.get_dipole_tddft()
    if driver.is_root :
        print('dipople:\n', dipole)
    driver.stop()
    # Restart the TDDFT
    driver = Driver(inputfile, comm, task = 'optical', iterative = True)
    # restart from 5
    driver.tddft_restart(istep=5)
    for i in range(5):
        driver.diagonalize()
    dipole = driver.get_dipole_tddft()
    if driver.is_root :
        print('dipople:\n', dipole)
    driver.stop()


if __name__ == "__main__":
    main()



import numpy as np
import qepy

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

fname = 'qe_in.in'

qepy.qepy_pwscf(fname, commf)

embed = qepy.qepy_common.embed_base()

qepy.qepy_electrons_scf(2, 0, embed)

nscf = qepy.control_flags.get_n_scf_steps()
conv_flag = bool(qepy.control_flags.get_conv_elec())
info = 'Converged {} at {} steps'.format(conv_flag, nscf)
qepy.qepy_mod.qepy_write_stdout(info)

# get density
nr = qepy.qepy_mod.qepy_get_grid()
nspin = qepy.lsda_mod.get_nspin()
rho = np.empty((np.prod(nr), nspin), order = 'F')
qepy.qepy_mod.qepy_get_rho(rho)
ncharge = rho.sum()*qepy.cell_base.get_omega()/np.prod(nr)
if comm is None or comm.rank == 0 : print('ncharge', ncharge, flush = True)

# qepy.qepy_calc_energies(embed)
etotal = embed.etotal

qepy.qepy_forces()
forces = qepy.force_mod.get_array_force().T

stress = np.ones((3, 3), order='F')
qepy.qepy_stress(stress)

qepy.punch('all')
qepy.qepy_stop_run(0, what = 'no')



#!/usr/bin/env python3
import qepy

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

from qepy.calculator import QEpyCalculator
import ase.io
from ase.io.trajectory import Trajectory
from ase import units
from ase.md.andersen import Andersen
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution

inputfile = 'qe_in.in'

calc = QEpyCalculator(comm = comm, inputfile = inputfile)

atoms = ase.io.read(inputfile, format='espresso-in')

atoms.set_calculator(calc)

T = 340
MaxwellBoltzmannDistribution(atoms, temperature_K = T, force_temp=True)

dyn = Andersen(atoms, 1.5 * units.fs, temperature_K = T, andersen_prob=0.02)

step = 0
interval = 1

def printenergy(a=atoms):
    global step, interval
    epot = a.get_potential_energy() / len(a)
    ekin = a.get_kinetic_energy() / len(a)
    if a.calc.is_root== 0 :
        print("Step={:<8d} Epot={:.5f} Ekin={:.5f} T={:.3f} Etot={:.5f}".format(
                step, epot, ekin, ekin / (1.5 * units.kB), epot + ekin), flush = True)
    step += interval


dyn.attach(printenergy, interval=1)

traj = Trajectory("md.traj", "w", atoms)
dyn.attach(traj.write, interval=1)

dyn.run(5)



import os
import sys
import re
import subprocess
import pathlib
import shutil

from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext

with open('qepy/__init__.py') as fd :
    lines = fd.read()
    __version__ = re.search('__version__ = "(.*)"', lines).group(1)
    __author__ = re.search('__author__ = "(.*)"', lines).group(1)
    __contact__ = re.search('__contact__ = "(.*)"', lines).group(1)
    __license__ = re.search('__license__ = "(.*)"', lines).group(1)

name = 'qepy'
description = "QEpy: Quantum ESPRESSO Python interface"
with open('README.md') as fh :
    long_description = fh.read()

fix_mpi4py = """# Fix the MPI_IN_PLACE and MKL
import sys, os
from ctypes import util, CDLL, RTLD_LOCAL, RTLD_GLOBAL
if 'mpi4py' in sys.modules :
    if hasattr(util, '_findLib_ld') and hasattr(util, '_get_soname') :
        mpilib = util._get_soname(util._findLib_ld('mpi'))
    else :
        mpilib = None
    mpilib = mpilib or util.find_library('mpi') or util.find_library('mpifort')
    try:
        CDLL(mpilib, RTLD_LOCAL | RTLD_GLOBAL)
    except Exception :
        pass
try:
    if hasattr(util, '_findLib_ld'):
        mkllib = os.path.basename(util._findLib_ld('mkl_rt'))
    else :
        mkllib = util.find_library('mkl_rt')
    CDLL(mkllib, RTLD_LOCAL | RTLD_GLOBAL)
except Exception :
    pass
# End fix
"""


class MakeBuild(build_ext):
    def run(self):
        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        topdir = './'
        build_args = []
        env = os.environ.copy()
        self.build_name = self.build_lib+ os.sep + name

        try:
            import multiprocessing as mp
            nprocs = max(mp.cpu_count()//2, 2)
        except ImportError:
            nprocs = 4
        build_args += ['-j', str(nprocs)]

        if env.get('qepydev', False):
            print("only remove *.so files", flush = True)
            for f in pathlib.Path(self.build_temp).glob('*.so'): os.remove(f)
        else :
            if os.path.exists(self.build_temp): shutil.rmtree(self.build_temp)

        if not os.path.exists(self.build_temp): os.makedirs(self.build_temp)

        makefiles = list(pathlib.Path(topdir + '/qepy/').glob('__*__.py')) + \
            list(pathlib.Path(topdir + '/install/').glob('*')) + \
            list(pathlib.Path(topdir + '/src/').glob('*'))

        qeversion = 'qe-' + env.get('qeversion', '6.5')

        for f in makefiles :
            if f.is_file():
                shutil.copy2(f, self.build_temp)
            else :
                if f.name.startswith('qe-') :
                    if f.name != qeversion : continue
                shutil.copytree(f, self.build_temp+os.sep+f.name)

        makefiles = list(pathlib.Path(self.build_temp).glob('**/*.f90'))
        for f in makefiles : shutil.copy2(f, self.build_temp)

        cwd = os.getcwd()
        os.chdir(self.build_temp)
        if env.get('ldau', 'no').lower() == 'yes' :
            sys.path.insert(0, './')
            from ldau.qepy_ldau_patch import ini2files
            ini2files('ldau/qepy_econf.ini')
        os.chdir(cwd)

        env['PYTHON'] = sys.executable

        # install the QE
        qedir = env.get('qedir', '')
        if not qedir :
            qedir = self.build_temp + '/q-e'
            qe_download = ["git", "clone", "-b", "qe-6.5", "--depth=1", "https://gitlab.com/QEF/q-e.git", qedir]
            subprocess.check_call(qe_download, env = env)
            if '-fPIC' not in env.get('CFLAGS', ''):
                env['CFLAGS'] = '-fPIC ' + env.get('CFLAGS', '')
            if '-fPIC' not in env.get('FFLAGS', ''):
                env['FFLAGS'] = '-fPIC ' + env.get('FFLAGS', '')
            if '-fPIC' not in env.get('try_foxflags', ''):
                env['try_foxflags'] = '-fPIC ' + env.get('try_foxflags', '')
            # blas and lapack
            if 'BLAS_LIBS' not in env : env['BLAS_LIBS'] = '-lblas'
            if 'LAPACK_LIBS' not in env : env['LAPACK_LIBS'] = '-llapack'
            qe_install_flags = env.get('QE_INSTALL_FLAGS', '').split('|')
            print('env', env)
            res = subprocess.run(["./configure"] + qe_install_flags, cwd=qedir, env = env, check=False, capture_output=True)
            stderr=res.stderr.decode()
            if 'error' in stderr:
                print('Some errors happened in configure...')
                subprocess.run(["cat", "install/config.log"], cwd=qedir, env = env, check = False)
                print(stderr)
                exit()
            subprocess.check_call(["make", "pwlibs"] + build_args, cwd=qedir, env = env)
            env['qedir'] = os.path.abspath(qedir)

        if env.get('tddft', 'no').lower() == 'yes' :
            subprocess.check_call(['make', '-f', 'Makefile.cetddft'] + build_args, cwd=self.build_temp, env = env)

        subprocess.check_call(['make', '-f', 'Makefile'] + build_args, cwd=self.build_temp, env = env)

        with open(self.build_temp + '/qepy/__init__.py', 'r+') as fh :
            lines = fh.readlines()
            fh.seek(0)
            first = True
            for line in lines :
                if first and '_qepy' in line :
                    fh.write(fix_mpi4py)
                    first = False
                fh.write(line)

        if not os.path.exists(self.build_lib): os.makedirs(self.build_lib)
        # if os.path.exists(self.build_name): shutil.rmtree(self.build_name)
        for f in pathlib.Path(self.build_temp + os.sep + name).glob('*'):
            shutil.copy2(f, self.build_name)
        for f in pathlib.Path(self.build_temp).glob('*.so'):
            shutil.copy2(f, self.build_lib + os.sep)


extensions_qepy = Extension(
        "qepy._qepy",
        sources = [],
        )

ext_modules = [extensions_qepy, ]

if __name__ == "__main__":
    setup(
            name=name,
            url='https://gitlab.com/shaoxc/qepy',
            description=description,
            version=__version__,
            # use_scm_version={'version_scheme': 'post-release'},
            author=__author__,
            author_email=__contact__,
            license=__license__,
            long_description=long_description,
            long_description_content_type='text/markdown',
            python_requires = '>=3.7',
            install_requires=[
                'setuptools_scm',
                'setuptools<=59.8.0',
                'numpy>=1.18.0',
                'f90wrap>=0.2.8',
                'importlib-metadata>=0.12;python_version<"3.8"'
                ],
            extras_require={
                'mpi': [
                    'mpi4py>=3.0.2',
                    ],
                },
            packages=find_packages('./'),
            ext_modules=ext_modules,
            # include_package_data=True,
            cmdclass = {"build_ext" : MakeBuild},
            classifiers=[
                'Development Status :: 3 - Alpha',
                'Intended Audience :: Science/Research',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
                'Programming Language :: Python :: 3.9',
                'Programming Language :: Python :: 3.10',
                'Topic :: Scientific/Engineering :: Chemistry',
                'Topic :: Scientific/Engineering :: Physics'
                ],
            zip_safe = False,
            )



import qepy

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    commf = comm.py2f()
except Exception:
    comm = None
    commf = None

fname = 'qe_in.in'
qepy.mp_global.mp_startup(my_world_comm=commf, start_images=True)
qepy.environment.environment_start('PWSCF')

qepy.read_input.read_input_file('PW', fname)

exit_status = qepy.run_pwscf()

qepy.laxlib_end()
qepy.stop_run(exit_status)
qepy.do_stop(exit_status)



import sys
import sphinx_rtd_theme
project = 'QEpy'
copyright = '2019, Pavanello Research Group'
author = 'Pavanello Research Group'
release = '0.0.1'

source_suffix = '.rst'
master_doc = 'index'

nbsphinx_execute = 'never'

extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.mathjax',
              'sphinx.ext.viewcode',
              'sphinx.ext.napoleon',
              'sphinx.ext.intersphinx',
              'sphinx.ext.autosummary',
              'nbsphinx',
              'sphinx_panels',
              ]

templates_path = ['templates']
exclude_patterns = ['build']
# exclude_patterns = ['build', 'source/tutorials/jupyter/*ipynb']

html_theme = 'sphinx_rtd_theme'
# html_theme = 'sphinx_bootstrap_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
# html_favicon = 'static/qepy.ico'
# html_logo = 'static/qepy.png'
html_style = 'custom.css'
html_static_path = ['static']
html_last_updated_fmt = '%A, %d %b %Y %H:%M:%S'

html_theme_options = {
    'prev_next_buttons_location': 'both',
}

latex_show_urls = 'inline'
latex_show_pagerefs = True
latex_documents = [('index', not True)]

smartquotes = False


#Add external links to source code
def linkcode_resolve(domain, info):
    print('info module', info)
    if domain != 'py' or not info['module']:
        return None

    filename = info['module'].replace('.', '/')+'.py'
    return "https ://gitlab.com/shaoxc/qepy/tree/master/%s" % filename



#!/usr/bin/env python3
import qepy
import time

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

from qepy.calculator import QEpyCalculator

inputfile = 'qe_in.in'

calc = QEpyCalculator(comm = comm, inputfile = inputfile)

get_potential_energy      = calc.get_potential_energy()
get_forces                = calc.get_forces()
get_stress                = calc.get_stress()
get_density               = calc.get_density()
get_bz_k_points           = calc.get_bz_k_points()
get_effective_potential   = calc.get_effective_potential()
get_eigenvalues           = calc.get_eigenvalues()
get_fermi_level           = calc.get_fermi_level()
get_ibz_k_points          = calc.get_ibz_k_points()
get_k_point_weights       = calc.get_k_point_weights()
get_magnetic_moment       = calc.get_magnetic_moment()
get_number_of_bands       = calc.get_number_of_bands()
get_number_of_grid_points = calc.get_number_of_grid_points()
get_number_of_spins       = calc.get_number_of_spins()
get_occupation_numbers    = calc.get_occupation_numbers()
get_pseudo_density        = calc.get_pseudo_density()
get_pseudo_wave_function  = calc.get_pseudo_wave_function()
get_spin_polarized        = calc.get_spin_polarized()
get_xc_functional         = calc.get_xc_functional()

time.sleep(0.1)

if calc.is_root :
    print('ncharge:', get_density.sum()*calc.atoms.get_volume()/get_density.shape[0] , flush = True)
    print("calc.get_potential_energy()           =" , get_potential_energy           , flush = True)
    print("calc.get_forces()[0]                  =" , get_forces[0]                  , flush = True)
    print("calc.get_stress()[0]                  =" , get_stress[0]                  , flush = True)
    print("calc.get_density()[0]                 =" , get_density[0]                 , flush = True)
    print("calc.get_bz_k_points()[:, 0]          =" , get_bz_k_points[:, 0]          , flush = True)
    print("calc.get_effective_potential()[0]     =" , get_effective_potential[0]     , flush = True)
    print("calc.get_eigenvalues()[0]             =" , get_eigenvalues[0]             , flush = True)
    print("calc.get_fermi_level()                =" , get_fermi_level                , flush = True)
    print("calc.get_ibz_k_points()[:, 0]         =" , get_ibz_k_points[:, 0]         , flush = True)
    print("calc.get_k_point_weights()[0]         =" , get_k_point_weights[0]         , flush = True)
    print("calc.get_magnetic_moment()            =" , get_magnetic_moment            , flush = True)
    print("calc.get_number_of_bands()            =" , get_number_of_bands            , flush = True)
    print("calc.get_number_of_grid_points()      =" , get_number_of_grid_points      , flush = True)
    print("calc.get_number_of_spins()            =" , get_number_of_spins            , flush = True)
    print("calc.get_occupation_numbers()[0]      =" , get_occupation_numbers[0]      , flush = True)
    print("calc.get_pseudo_density()[0]          =" , get_pseudo_density[0]          , flush = True)
    print("calc.get_pseudo_wave_function()[0, 0] =" , get_pseudo_wave_function[0, 0] , flush = True)
    print("calc.get_spin_polarized()             =" , get_spin_polarized             , flush = True)
    print("calc.get_xc_functional()              =" , get_xc_functional              , flush = True)

time.sleep(0.1)



from qepy.driver import Driver

try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
except Exception:
    comm = None

def main():
    inputfile = './qe_in.in'
    driver = Driver(inputfile, comm)
    driver.scf()
    converged = driver.check_convergence()
    #
    energy = driver.get_energy()
    forces = driver.get_forces()
    stress = driver.get_stress()
    if driver.is_root :
        print('converged :\n', converged)
        print('energy :\n', energy)
        print('forces :\n', forces)
        print('stress :\n', stress)
    driver.stop()


if __name__ == "__main__":
    main()
